"""Add the Area model

Revision ID: 54db49335a2
Revises: 2367ebea90d
Create Date: 2014-04-20 13:43:33.248879

"""

# revision identifiers, used by Alembic.
revision = '54db49335a2'
down_revision = '2367ebea90d'

from datetime import datetime

from alembic import op
import sqlalchemy as sa

connection = op.get_bind()


def upgrade():
    op.create_table(
        'area',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('name', sa.String(length=20), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_index('ix_area_created_at', 'area', ['created_at'], unique=False)
    op.create_index('ix_area_name', 'area', ['name'], unique=True)
    op.add_column('device', sa.Column('area_id', sa.Integer()))

    # Create two hybrid tables that represent the new table and the old
    # device table with the new area_id column.
    area_migrator = sa.Table(
        'area',
        sa.MetaData(),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('name', sa.String(length=20), nullable=True)
    )

    device_migrator = sa.Table(
        'device',
        sa.MetaData(),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('name', sa.String(length=20), nullable=True),
        sa.Column('device_type', sa.Integer(), nullable=True),
        sa.Column('device_sub_type', sa.Integer(), nullable=True),
        sa.Column('device_id', sa.String(length=20), nullable=False),
        sa.Column('area_id', sa.Integer())
    )

    for device in connection.execute(device_migrator.select()):

        name = device.name

        connection.execute(
            area_migrator.insert()
            .values(name=name,
                    created_at=datetime.utcnow()))

        # After inserting, grab it back. We need to get the ID.
        r = connection.execute(
            area_migrator
            .select().where(
                area_migrator.c.name == name,
            ))

        area = r.first()

        # Update the data_point table to populate the new device_series_id.
        connection.execute(
            device_migrator.update().where(
                device_migrator.c.id == device.id
            ).values(
                area_id=area['id']
            )
        )

    # Ok, now its fully populated. Disable nulls on the col. If there are still
    # nulls we will get an error here, so its a good sanity check.
    op.alter_column('device', 'area_id', nullable=False)

    # Finally drop the old stuff as per the alembic generated code.
    op.drop_column('device', 'name')


def downgrade():
    op.create_index('ix_device_name', 'device', ['name'], unique=True)
    op.add_column('device', sa.Column('name', sa.VARCHAR(length=20),
                  nullable=True))
    op.drop_column('device', 'area_id')
    op.drop_index('ix_area_name', table_name='area')
    op.drop_index('ix_area_created_at', table_name='area')
    op.drop_table('area')



from datetime import datetime

from tests.base import BaseTestCase

from home.ts.models import Area, Device, Series, DataPoint, DeviceSeries


class ValuesResourceTestCase(BaseTestCase):

    def setUp(self):
        super().setUp()

        now = datetime.now()

        self.series = Series(name="Test series")
        self.area = Area(name="livingroom")
        self.device = Device(82, 1, '0xAAAA', area=self.area)
        self.device_series = DeviceSeries(self.device, self.series)
        self.value = DataPoint(self.device_series, 10, created_at=now)

        self.insert(self.device)
        self.insert(self.series)
        self.insert(self.value)

    def test_dashboard(self):

        r = self.client.get('/')

        self.assertEqual(200, r.status_code)

    def test_area(self):

        r = self.client.get('/areas/livingroom/')

        self.assertEqual(200, r.status_code)

    def test_device(self):

        r = self.client.get('/device/1/')

        self.assertEqual(404, r.status_code)



from home import manager


@manager.option('--device', help='Serial device.')
def collect(device):
    "Start collecting data from the given serial device"
    from home.collect import run
    run(device)


def main():
    manager.run()

if __name__ == '__main__':
    main()



from home import app
from home.collect.handlers import RecordingHander, LoggingHandler
from home.collect.loop import collect


elec_handler = RecordingHander({
    'current watts': 'current_watts',
    'total watts': 'total_watts'
})

temp_humidity_handler = RecordingHander({
    'temperature': 'temperature',
    'humidity': 'humidity'
})

logging_handler = LoggingHandler()


def run(device):

    with app.app_context():
        collect(device)



"""indexes

Revision ID: 2584ea13b8a
Revises: 25768ce6905
Create Date: 2014-04-20 09:23:42.942280

"""

# revision identifiers, used by Alembic.
revision = '2584ea13b8a'
down_revision = '25768ce6905'

from alembic import op


def upgrade():
    op.create_index('ix_data_point_device_series_id', 'data_point',
                    ['device_series_id'], unique=False)
    op.create_index('ix_data_point_value', 'data_point', ['value'],
                    unique=False)
    op.create_index('ix_device_created_at', 'device', ['created_at'],
                    unique=False)
    op.create_index('ix_device_device_sub_type', 'device', ['device_sub_type'],
                    unique=False)
    op.create_index('ix_device_device_type', 'device', ['device_type'],
                    unique=False)
    op.create_index('ix_device_name', 'device', ['name'], unique=True)
    op.create_index('ix_device_series_created_at', 'device_series',
                    ['created_at'], unique=False)
    op.create_index('ix_graph_name', 'graph', ['name'], unique=True)
    op.create_index('ix_series_created_at', 'series', ['created_at'],
                    unique=False)
    op.create_index('ix_series_name', 'series', ['name'], unique=True)


def downgrade():
    op.drop_index('ix_series_name', table_name='series')
    op.drop_index('ix_series_created_at', table_name='series')
    op.drop_index('ix_graph_name', table_name='graph')
    op.drop_index('ix_device_series_created_at', table_name='device_series')
    op.drop_index('ix_device_name', table_name='device')
    op.drop_index('ix_device_device_type', table_name='device')
    op.drop_index('ix_device_device_sub_type', table_name='device')
    op.drop_index('ix_device_created_at', table_name='device')
    op.drop_index('ix_data_point_value', table_name='data_point')
    op.drop_index('ix_data_point_device_series_id', table_name='data_point')



from logging.config import dictConfig
from os import path
from sys import stdout

SQLALCHEMY_DATABASE_URI = 'postgresql://home:home@localhost:5432/home'

PROJECT_PATH = path.dirname(path.realpath(__file__))
TEMPLATE_FOLDER = path.join(PROJECT_PATH, 'dash', 'templates')
STATIC_FOLDER = path.join(PROJECT_PATH, 'dash', 'static')

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s %(levelname)-8s %(name)-35s %(message)s'
        },
    },
    'handlers': {
        'console': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'stream': stdout,
            'formatter': 'standard'
        },
        'file': {
            'level': 'WARNING',
            'class': 'logging.handlers.RotatingFileHandler',
            'formatter': 'standard',
            'filename': path.join(PROJECT_PATH, "home.log"),
            'maxBytes': 10 * 1024 * 1024,
        },
        'file-full': {
            'level': 'DEBUG',
            'class': 'logging.handlers.RotatingFileHandler',
            'formatter': 'standard',
            'filename': path.join(PROJECT_PATH, "home-full.log"),
            'maxBytes': 10 * 1024 * 1024,
        },
    },
    'loggers': {
        'rfxcom': {
            'handlers': ['console', 'file', 'file-full'],
            'propagate': True,
            'level': 'DEBUG',
        },
        'home': {
            'handlers': ['console', 'file', 'file-full'],
            'propagate': True,
            'level': 'DEBUG',
        }
    },
}

DEBUG_TB_INTERCEPT_REDIRECTS = False

dictConfig(LOGGING)


PACKET_HANDLERS = {
    'rfxcom.protocol.Status': 'home.collect.logging_handler',
    'rfxcom.protocol.Elec': 'home.collect.elec_handler',
    'rfxcom.protocol.TempHumidity': 'home.collect.temp_humidity_handler',
    '*': 'home.collect.logging_handler',
}



"""M2M Change

Revision ID: 43e49e93f5d
Revises: 59333101ef2
Create Date: 2014-04-10 15:36:40.970441

"""

# revision identifiers, used by Alembic.
revision = '43e49e93f5d'
down_revision = '59333101ef2'

from datetime import datetime

from alembic import op
import sqlalchemy as sa

connection = op.get_bind()


def upgrade():

    # Create the device_series table as per the default alembic generated code.
    op.create_table(
        'device_series',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('series_id', sa.Integer(), nullable=False),
        sa.Column('device_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['device_id'], ['device.id'], ),
        sa.ForeignKeyConstraint(['series_id'], ['series.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('series_id', 'device_id', name='_series_device_uc')
    )

    # Add the relation to the new table in data_point, but remove the
    # nullable=False parameter. We need to allow nulls initially.
    op.add_column('data_point', sa.Column('device_series_id', sa.Integer()))

    # Create two hybrid tables that represent the new table and the old
    # data_point table with the new relationship column.
    data_point_migrator = sa.Table(
        'data_point',
        sa.MetaData(),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('value', sa.Numeric(), nullable=False),
        sa.Column('series_id', sa.Integer(), nullable=True),
        sa.Column('device_id', sa.Integer(), nullable=True),
        sa.Column('device_series_id', sa.Integer(), nullable=True),
    )

    device_series_migrator = sa.Table(
        'device_series',
        sa.MetaData(),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('series_id', sa.Integer(), nullable=True),
        sa.Column('device_id', sa.Integer(), nullable=True),
    )

    # We know the unique constraint on device_services, so store them as we go
    # to make things a bit faster.
    seen_device_series = {}

    # Iterate through all of the data points. We want to extract the relation
    # on data_point to series and device and put it in the new device_series
    # table.
    for data_point in connection.execute(data_point_migrator.select()):

        # Pull the data from data_point.
        s_id = data_point.series_id
        d_id = data_point.device_id

        # Insert into device_series if it doesn't already exist.
        if (s_id, d_id) not in seen_device_series:

            # Simple insert, just use utcnow - we don't care much about that
            # it will be more useful for future inserts.
            connection.execute(
                device_series_migrator.insert()
                .values(series_id=s_id, device_id=d_id,
                        created_at=datetime.utcnow()))

            # After inserting, grab it back. We need to get the ID.
            r = connection.execute(
                device_series_migrator
                .select().where(
                    device_series_migrator.c.series_id == s_id,
                ).where(
                    device_series_migrator.c.device_id == d_id
                ))

            # cache the result for future iterations.
            seen_device_series[(s_id, d_id)] = dict(r.first())

        device_series = seen_device_series[(s_id, d_id)]

        # Update the data_point table to populate the new device_series_id.
        connection.execute(
            data_point_migrator.update().where(
                data_point_migrator.c.id == data_point.id
            ).values(
                device_series_id=device_series['id']
            )
        )

    # Ok, now its fully populated. Disable nulls on the col. If there are still
    # nulls we will get an error here, so its a good sanity check.
    op.alter_column('data_point', 'device_series_id', nullable=False)

    # Finally, drop the two original columns as per the end of the original
    # alembic code.
    op.drop_column('data_point', 'series_id')
    op.drop_column('data_point', 'device_id')


def downgrade():
    op.add_column(
        'data_point', sa.Column('device_id', sa.INTEGER(), nullable=True))
    op.add_column(
        'data_point', sa.Column('series_id', sa.INTEGER(), nullable=True))
    op.drop_column('data_point', 'device_series_id')
    op.drop_table('device_series')



from datetime import datetime

from tests.base import BaseTestCase

from home.ts.models import Area, Series, Device, DataPoint, DeviceSeries


class SeriesModelTestCase(BaseTestCase):

    def setUp(self):
        super().setUp()

        self.series = Series(name="Series")
        self.insert(self.series)

    def test_repr(self):

        repr(self.series)


class DeviceModelTestCase(BaseTestCase):

    def setUp(self):
        super().setUp()

        self.area = Area(name="livingroom")
        self.device = Device(82, 1, '0xAAAA', area=self.area)
        self.insert(self.device)

    def test_repr(self):

        repr(self.device)


class DataPointModelTestCase(BaseTestCase):

    def setUp(self):
        super().setUp()

        now = datetime.now()

        self.series = Series(name="series_a")
        self.area = Area(name="livingroom")
        self.device = Device(82, 1, '0xAAAA', area=self.area)
        self.device_series = DeviceSeries(self.device, self.series)
        self.data_point = DataPoint(self.device_series, 10, created_at=now)

        self.insert(self.device)
        self.insert(self.series)
        self.insert(self.data_point)

    def test_repr(self):

        repr(self.data_point)

    def test_record(self):

        DataPoint.record(self.series, self.device, 100)



from asyncio import get_event_loop

from rfxcom.transport import AsyncioTransport

from home.config import PACKET_HANDLERS
from home.collect.handlers import load_handlers


def collect(dev_name, callbacks=None):

    if callbacks is None:
        callbacks = dict(load_handlers(PACKET_HANDLERS))

    loop = get_event_loop()

    try:

        AsyncioTransport(dev_name, loop, callbacks=callbacks)
        loop.run_forever()

    finally:
        loop.close()



from sys import version_info
from unittest.mock import Mock, patch, ANY
from unittest import skipIf

from tests.base import BaseTestCase

from rfxcom.protocol import Status

from home.collect import logging_handler
from home.collect.handlers import (BaseHandler, LoggingHandler,
                                   RecordingHander, load_handlers)
from home.exceptions import HandlerConfigError


def mock_packet(extra=None):

    p = Mock()
    p.raw = bytearray(b'\x11\x5A\x01\x00\x2E\xB2\x03\x00\x00')
    p.data = {
        'id': 'test',
        'packet_type': 33,
        'sub_type': 1,
    }

    if extra is not None:
        p.data.update(extra)

    return p


class HandlerLoadingTestCase(BaseTestCase):

    def test_load_handlers_empty(self):

        result = load_handlers({})

        self.assertEqual(result, {})

    def test_load_handlers_duplicate(self):

        with self.assertRaises(HandlerConfigError):
            load_handlers({
                'rfxcom.protocol.Status': 'home.collect.logging_handler',
                Status: 'home.collect.logging_handler',
            })

    def test_load_handlers_bad_path(self):

        with self.assertRaises(ImportError):
            load_handlers({
                'rfxcom.protocol..Status': 'home.collect.logging_handler'
            })

    def test_load_handlers_wildcard(self):

        result = load_handlers({
            '*': 'home.collect.logging_handler',
        })

        self.assertEqual(result, {
            '*': logging_handler
        })

    def test_load_handlers_all_objects(self):

        h = lambda x: x

        result = load_handlers({
            Status: h,
        })

        self.assertEqual(result, {
            Status: h
        })


class BaseHandlerTestCase(BaseTestCase):

    def setUp(self):

        super().setUp()
        self.packet = mock_packet()
        self.handler = BaseHandler()

    def test_logger(self):

        self.assertEqual(self.handler.log.name, 'home.collect.BaseHandler')

    def test_format_packet(self):

        formatted = self.handler.format_packet(self.packet.raw)
        expected = '0x11 0x5a 0x01 0x00 0x2e 0xb2 0x03 0x00 0x00'

        self.assertEqual(formatted, expected)


class LoggingHandlerTestCase(BaseTestCase):

    def setUp(self):

        super().setUp()
        self.packet = mock_packet()
        self.handler = LoggingHandler()
        self.log_name = 'home.collect.LoggingHandler'

    def test_logger(self):

        self.assertEqual(self.handler.log.name, self.log_name)

    @skipIf(version_info < (3, 4), "Python 3.4 required.")
    def test_packet_logged(self):

        with self.assertLogs(self.log_name, level='INFO') as cm:

            self.handler(self.packet)

            self.assertEqual(cm.output, [
                'WARNING:home.collect.LoggingHandler:Ignoring packet: 0x11 '
                '0x5a 0x01 0x00 0x2e 0xb2 0x03 0x00 0x00'
            ])


class RecordingHanderTestCase(BaseTestCase):

    def setUp(self):

        super().setUp()
        self.packet = mock_packet({
            'current_watts': 650,
            'total_watts': 100000
        })
        self.handler = RecordingHander({
            'electricity': 'current_watts',
        })
        self.log_name = 'home.collect.RecordingHander'

    def test_logger(self):

        n = self.handler.log.name
        self.assertEqual(n, 'home.collect.RecordingHander')

    @patch('home.ts.models.DataPoint.record')
    def test_record(self, mock_record):

        self.handler(self.packet)

        mock_record.assert_called_once_with(ANY, ANY, 650)

    @skipIf(version_info < (3, 4), "Python 3.4 required.")
    @patch('home.ts.models.DataPoint.record')
    def test_record_missing(self, mock_record):

        self.handler = RecordingHander({
            'electricity': 'not_a_correct_attribute',
        })

        with self.assertLogs(self.log_name, level='ERROR') as cm:
            self.handler(self.packet)
            self.assertEqual(cm.output, [
                "ERROR:home.collect.RecordingHander:Failed to find "
                "not_a_correct_attribute in packet. Key list: ['current_watts'"
                ", 'id', 'packet_type', 'sub_type', 'total_watts']"
            ])

        assert not mock_record.called, 'method should not have been called'



"""Initial

Revision ID: 59333101ef2
Revises: None
Create Date: 2014-04-10 14:18:20.235516

"""

# revision identifiers, used by Alembic.
revision = '59333101ef2'
down_revision = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'series',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('name', sa.String(length=20), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name')
    )
    op.create_table(
        'device',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('name', sa.String(length=20), nullable=True),
        sa.Column('device_type', sa.Integer(), nullable=True),
        sa.Column('device_sub_type', sa.Integer(), nullable=True),
        sa.Column('device_id', sa.String(length=20), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('device_id'),
        sa.UniqueConstraint('name')
    )
    op.create_table(
        'data_point',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('value', sa.Numeric(), nullable=False),
        sa.Column('series_id', sa.Integer(), nullable=True),
        sa.Column('device_id', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(['device_id'], ['device.id'], ),
        sa.ForeignKeyConstraint(['series_id'], ['series.id'], ),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('data_point')
    op.drop_table('device')
    op.drop_table('series')



from datetime import timedelta
from functools import partial
from itertools import groupby

AGGREGATORS = {
    'delta': lambda dps: max(dps) - min(dps),
    'max': max,
    'mean': lambda dps: round(sum(dps) / min(len(dps), 1), 2),
    'min': min,
}


class AggregatedResultSet:

    def __init__(self, group, duration, value):

        self.group = group
        self.duration = duration
        self.value = value

    def as_dict(self):
        return {
            'created_at': self.group,
            'value': self.value
        }


def round_datetime(data_point, round_by):
    seconds = (data_point.created_at - data_point.created_at.min).seconds
    rounding = (seconds + round_by / 2) // round_by * round_by
    d = timedelta(0, rounding-seconds, -data_point.created_at.microsecond)
    return data_point.created_at + d


def group_values(values, aggregator_functions):

    seconds = 60 * 60

    key = partial(round_datetime, round_by=seconds)

    results = [[] for _ in aggregator_functions]

    for group, group_values in groupby(values, key=key):

        group_values = list(group_values)

        for i, aggregator_function in enumerate(aggregator_functions):
            f = AGGREGATORS[aggregator_function]
            value = f([dp.value for dp in group_values])

            results[i].append(AggregatedResultSet(group=group,
                              duration=seconds, value=value))

    return results


FUNCTIIONS = {
    'minmax': partial(group_values, aggregator_functions=['min', 'max', ]),
    'delta': partial(group_values, aggregator_functions=['delta', ])
}


def get_method(graph_model):

    if graph_model is None:
        return

    f = FUNCTIIONS.get(graph_model.aggregator)

    return f



from functools import wraps

from flask import Blueprint, request, jsonify, abort
from flask.views import MethodView
from werkzeug.exceptions import BadRequest

from home.ts import graph
from home.ts.models import Area, Device, Series, DataPoint, DeviceSeries

api = Blueprint('Dashboard API', __name__)


def register_api(view, endpoint, url):
    view_func = view.as_view(endpoint)

    api.add_url_rule(url, defaults={'resource_id': None, 'name': None},
                     view_func=view_func, methods=['GET', ])

    api.add_url_rule(url, view_func=view_func, methods=['POST', ])

    api.add_url_rule('%s<int:resource_id>' % (url), defaults={'name': None},
                     view_func=view_func,
                     methods=['GET', 'PUT', 'DELETE'])

    api.add_url_rule('%s<string:name>' % (url), defaults={'resource_id': None},
                     view_func=view_func,
                     methods=['GET', 'PUT', 'DELETE'])


def kwarg_json_query(f):

    @wraps(f)
    def wrapper(*args, **kwargs):

        try:
            json = request.get_json(force=True)
        except BadRequest:
            json = {}

        if isinstance(json, dict):
            kwargs['query'] = json
        else:
            raise Exception("Invalid JSON")

        return f(*args, **kwargs)

    return wrapper


class JSONResource(MethodView):

    def jsonify_qs(self, result_sets, **kwargs):

        r = []

        for result_set in result_sets:
            r.append({
                'values': [i.as_dict() for i in result_set],
                'count': len(list(result_set)),
            })

        return jsonify(data=r, **kwargs)


class Resource(JSONResource):

    page_size = 1000

    def get(self, resource_id=None, name=None):

        if resource_id is not None:
            resource = self.model.query.filter_by(id=resource_id).first()
            if not resource:
                abort(404)
            return jsonify(resource.as_dict())

        if name is not None:
            resource = self.model.query.filter_by(name=name).first()
            return jsonify(resource.as_dict())

        page = max(int(request.args.get('page', '0')), 1)

        results = self.model.query.order_by(
            self.model.created_at.desc())
        offset = (page - 1) * self.page_size
        results = results.limit(self.page_size).offset(offset)
        return self.jsonify_qs([results, ], page=page)


class DevicesResource(Resource):
    model = Device


class SeriesResource(Resource):
    model = Series


class DeviceSeriesResource(Resource):
    model = DeviceSeries


class ValuesResource(Resource):
    model = DataPoint


class AreasResource(Resource):
    model = Area


class GraphResource(ValuesResource):

    @kwarg_json_query
    def post(self, query=None):

        graph_func = None
        device_series_id = None

        if query is None:
            abort(400)

        qs = DataPoint.query.order_by(DataPoint.created_at.desc())

        if 'device_id' in query and 'series_id' in query:

            ds = DeviceSeries.query.filter_by(
                device_id=query['device_id'],
                series_id=query['series_id'],
            ).first()

            if 'graph' not in query:
                graph_func = graph.get_method(ds.series.graph)

            if ds is not None:
                device_series_id = ds.id

        if 'device_series_id' in query:
            device_series_id = query['device_series_id']

        if device_series_id is not None:
            qs = qs.filter_by(device_series_id=device_series_id)

        if 'start' in query and 'end' in query:
            start = query['start']
            end = query['end']
            qs = qs.filter(DataPoint.created_at.between(start, end))

        qs = qs.limit(10000)

        if graph_func is not None:
            qs = graph_func(qs)
        else:
            qs = [qs, ]

        return self.jsonify_qs(qs)


register_api(AreasResource, 'areas', '/areas/')
register_api(DeviceSeriesResource, 'device_series', '/device_series/')
register_api(DevicesResource, 'devices', '/devices/')
register_api(GraphResource, 'graph', '/graph/')
register_api(SeriesResource, 'series', '/series/')
register_api(ValuesResource, 'values', '/values/')



__version__ = '0.1.0'

from flask import Flask
from flask.ext.admin import Admin
from flask.ext.admin.contrib.sqla import ModelView
from flask.ext.migrate import Migrate, MigrateCommand
from flask.ext.script import Manager

from home.config import TEMPLATE_FOLDER, STATIC_FOLDER
from home.dash.api import api
from home.dash.web import web
from home.ts.models import db, Graph, Series, Device, DeviceSeries, Area


def create_app(config=None):

    app = Flask(__name__, static_folder=STATIC_FOLDER,
                template_folder=TEMPLATE_FOLDER)

    app.config['SECRET_KEY'] = 'ssh, its a secret.'

    app.config.from_object('home.config')
    app.config.from_object(config)

    app.register_blueprint(web)
    app.register_blueprint(api, url_prefix='/api')

    db.init_app(app)

    Migrate(app, db, directory='home/migrations')

    admin = Admin(app)
    admin.add_view(ModelView(Area, db.session))
    admin.add_view(ModelView(Device, db.session))
    admin.add_view(ModelView(Series, db.session))
    admin.add_view(ModelView(DeviceSeries, db.session))
    admin.add_view(ModelView(Graph, db.session))

    return app

app = create_app()

manager = Manager(app)
manager.add_command('db', MigrateCommand)



from logging import getLogger

from home import db
from home.exceptions import HandlerConfigError
from home.ts.models import Device, Series, DataPoint


def importer(dotted_path):
    module, object_ = dotted_path.rsplit('.', 1)
    try:
        mod = __import__(module, fromlist=[object_, ])
    except ImportError as e:
        raise ImportError("Failed to load {0}".format(dotted_path)) from e
    return getattr(mod, object_)


def load_handlers(handler_mapping):
    """
    Given a dictionary mapping which looks like the following, import the
    objects based on the dotted path and yield the packet type and handler as
    pairs.

    If the special string '*' is passed, don't process that, pass it on as it
    is a wildcard.

    If an non-string object is given for either packet or handler (key or
    value) assume these are the objects to use and yield them.

    {
        'rfxcom.protocol.Status': 'home.collect.logging_handler',
        'rfxcom.protocol.Elec': 'home.collect.elec_handler',
        'rfxcom.protocol.TempHumidity': 'home.collect.temp_humidity_handler',
        '*': 'home.collect.logging_handler'
    }
    """

    handlers = {}

    for packet_type, handler in handler_mapping.items():

        if packet_type == '*':
            Packet = packet_type
        elif isinstance(packet_type, str):
            Packet = importer(packet_type)
        else:
            Packet = packet_type

        if isinstance(handler, str):
            Handler = importer(handler)
        else:
            Handler = handler

        if Packet in handlers:
            raise HandlerConfigError(
                "Handler already provided for packet %s" % Packet)

        handlers[Packet] = Handler

    return handlers


class BaseHandler:

    def __init__(self):

        self.log = getLogger('home.collect.%s' % self.__class__.__name__)

    def format_packet(self, pkt):

        return " ".join("0x{0:02x}".format(x) for x in pkt)


class LoggingHandler(BaseHandler):

    def __call__(self, packet):

        self.log.warning(
            "Ignoring packet: {0}".format(self.format_packet(packet.raw)))


class RecordingHander(BaseHandler):

    def __init__(self, mapping):
        super().__init__()
        self.mapping = mapping

    def __call__(self, packet):

        device = Device.get_or_create(packet.data['packet_type'],
                                      packet.data['sub_type'],
                                      packet.data['id'])

        for series_name, value_name in self.mapping.items():

            series = Series.get_or_create(name=series_name)
            db.session.commit()
            try:
                val = packet.data[value_name]
            except KeyError:
                self.log.error("Failed to find %s in packet. Key list: %r" % (
                    value_name, sorted(packet.data.keys())))
                return

            id_ = packet.data.get('id')
            self.log.info("ID=%s, %s=%s" % (id_, series_name, val))
            DataPoint.record(series, device, val)

        db.session.commit()



from os import environ

from flask.ext.testing import TestCase

from home import create_app
from home.ts.models import db

if environ.get('TRAVIS') == 'true':
    _DATABASE_URI = 'postgresql://postgres@localhost:5432/test_home'
else:
    _DATABASE_URI = 'postgresql://home:home@localhost:5432/test_home'


class BaseTestCase(TestCase):

    SQLALCHEMY_DATABASE_URI = _DATABASE_URI
    TESTING = True

    def create_app(self):

        app = create_app(self)

        self.assertEqual(app.config['SQLALCHEMY_DATABASE_URI'], _DATABASE_URI)

        return app

    def setUp(self):

        self.db = db
        self.db.create_all()

    def insert(self, model):
        self.db.session.add(model)
        self.db.session.commit()

    def tearDown(self):

        self.db.session.remove()
        self.db.drop_all()



"""Add graph description

Revision ID: 2367ebea90d
Revises: 2584ea13b8a
Create Date: 2014-04-20 11:21:00.657168

"""

# revision identifiers, used by Alembic.
revision = '2367ebea90d'
down_revision = '2584ea13b8a'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('graph', sa.Column('description', sa.Text(), nullable=True))


def downgrade():
    op.drop_column('graph', 'description')



"""Index DataPoint.created_at

Revision ID: 1afb70e89d2
Revises: 43e49e93f5d
Create Date: 2014-04-12 15:36:18.025078

"""

# revision identifiers, used by Alembic.
revision = '1afb70e89d2'
down_revision = '43e49e93f5d'

from alembic import op


def upgrade():
    op.create_index('ix_data_point_created_at', 'data_point', ['created_at'],
                    unique=False)


def downgrade():
    op.drop_index('ix_data_point_created_at', table_name='data_point')



import ast
import codecs
import os

from setuptools import setup, find_packages


class VersionFinder(ast.NodeVisitor):
    def __init__(self):
        self.version = None

    def visit_Assign(self, node):
        if getattr(node.targets[0], 'id', None) == '__version__':
            self.version = node.value.s


def read(*parts):
    filename = os.path.join(os.path.dirname(__file__), *parts)
    with codecs.open(filename, encoding='utf-8') as fp:
        return fp.read()


def find_version(*parts):
    finder = VersionFinder()
    finder.visit(ast.parse(read(*parts)))
    return finder.version


setup(
    name="home",
    version=find_version("home", "__init__.py"),
    url='https://github.com/d0ugal/home',
    license='BSD',
    description="Home automation project by d0ugal.",
    long_description=read('README.rst'),
    author='Dougal Matthews',
    author_email='dougal@dougalmatthews.com',
    packages=find_packages(exclude=["tests*"]),
    install_requires=[
        'alembic==0.6.4',
        'Flask-Admin==1.0.7',
        'Flask-Migrate==1.2.0',
        'Flask-SQLAlchemy==1.0',
        'Flask==0.10.1',
        'psycopg2==2.5.2',
        'rfxcom==0.0.1',
        'simplejson==3.3.3',
        'SQLAlchemy==0.9.3',
    ],
    include_package_data=True,
    platforms='any',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    entry_points={
        'console_scripts': [
            'home = home.__main__:main'
        ]
    },
    zip_safe=False,
)




def get_or_create(model, **kwargs):

    from home import db

    instance = model.query.filter_by(**kwargs).first()
    if instance:
        return instance
    else:
        instance = model(**kwargs)
        db.session.add(instance)
        return instance



"""Graph model

Revision ID: 25768ce6905
Revises: 1afb70e89d2
Create Date: 2014-04-20 09:17:23.956372

"""

# revision identifiers, used by Alembic.
revision = '25768ce6905'
down_revision = '1afb70e89d2'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'graph',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(length=20), nullable=False),
        sa.Column('aggregator', sa.String(length=20), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name')
    )
    op.drop_index('data_point_created_at_idx', table_name='data_point')
    op.drop_index('data_point_value_idx', table_name='data_point')
    op.add_column('series', sa.Column('graph_id', sa.Integer(), nullable=True))


def downgrade():
    op.drop_column('series', 'graph_id')
    op.create_index('data_point_value_idx', 'data_point', ['value'],
                    unique=False)
    op.create_index('data_point_created_at_idx', 'data_point', ['created_at'],
                    unique=False)
    op.drop_table('graph')



from datetime import datetime

from sqlalchemy import (Column, Numeric, Integer, String, ForeignKey, DateTime,
                        UniqueConstraint, Text)
from sqlalchemy.orm import relationship, backref
from flask.ext.sqlalchemy import SQLAlchemy

from home.util import get_or_create

db = SQLAlchemy()


class SerialiseMixin:

    def as_dict(self):
        return {c.name: getattr(self, c.name) for c in self.__table__.columns}


class DataPoint(db.Model, SerialiseMixin):
    __tablename__ = 'data_point'

    id = Column(Integer, primary_key=True)
    created_at = Column(DateTime, nullable=False, index=True)
    value = Column(Numeric, nullable=False, index=True)
    device_series_id = Column(Integer, ForeignKey('device_series.id'),
                              index=True, nullable=False)

    device_series = relationship(
        "DeviceSeries", order_by=created_at.desc(),
        backref=backref('data_points'))

    def __init__(self, device_series, value, created_at=None):
        self.created_at = datetime.utcnow()
        super().__init__()
        self.device_series = device_series
        self.value = value
        self.created_at = created_at

    @classmethod
    def record(cls, series, device, value, created_at=None):
        ds = DeviceSeries.get_or_create(device=device, series=series)
        data_point = DataPoint(device_series=ds, value=value)
        if created_at is None:
            data_point.created_at = datetime.utcnow()
        db.session.add(data_point)
        return data_point

    def __repr__(self):
        return "<Data Point(%s, value=%s, created_at=%s)>" % (
            self.device_series, self.value, self.created_at)


class DeviceSeries(db.Model, SerialiseMixin):
    __tablename__ = 'device_series'

    id = Column(Integer, primary_key=True)
    created_at = Column(DateTime, nullable=False, index=True)
    series_id = Column(Integer, ForeignKey('series.id'), nullable=False)
    device_id = Column(Integer, ForeignKey('device.id'), nullable=False)

    device = relationship(
        "Device", backref=backref('device_series'), lazy='joined')
    series = relationship(
        "Series", backref=backref('device_series'), lazy='joined')

    __table_args__ = (
        UniqueConstraint('series_id', 'device_id', name='_series_device_uc'),
    )

    def __init__(self, device, series):
        super().__init__()
        self.created_at = datetime.utcnow()
        self.series = series
        self.device = device

    @property
    def latest_reading(self):
        return DataPoint.query\
            .filter_by(device_series=self)\
            .order_by(DataPoint.created_at.desc())\
            .limit(1).first()

    @classmethod
    def get_or_create(cls, **kwargs):
        return get_or_create(cls, **kwargs)

    def __repr__(self):
        return "DeviceSeries(device_id=%s, series_id=%s)" % (
            self.device_id, self.series_id)


class Graph(db.Model, SerialiseMixin):

    __tablename__ = 'graph'

    id = Column(Integer, primary_key=True)
    name = Column(String(20), nullable=False, index=True, unique=True)
    description = Column(Text)
    aggregator = Column(String(20), nullable=False)

    def __repr__(self):
        return "Graph(name=%s)" % (self.name, )


class Series(db.Model, SerialiseMixin):
    __tablename__ = 'series'

    id = Column(Integer, primary_key=True)
    created_at = Column(DateTime, index=True, nullable=False)
    name = Column(String(20), nullable=False, index=True, unique=True)
    graph_id = Column(Integer, ForeignKey('graph.id'))

    graph = relationship(
        "Graph", backref=backref('series'), lazy='joined', uselist=False)

    def __init__(self, name):
        super().__init__()
        self.created_at = datetime.utcnow()
        self.name = name

    @classmethod
    def get_or_create(cls, **kwargs):
        r = get_or_create(cls, **kwargs)
        return r

    def __repr__(self):
        return "Series(name=%r)" % (self.name)


class Area(db.Model, SerialiseMixin):
    __tablename__ = 'area'

    id = Column(Integer, primary_key=True)
    created_at = Column(DateTime, index=True, nullable=False)
    name = Column(String(20), nullable=True, index=True, unique=True)

    def __init__(self, name):
        super().__init__()
        self.created_at = datetime.utcnow()
        self.name = name

    def __repr__(self):
        return "Area(name=%r)" % (self.name)


class Device(db.Model, SerialiseMixin):
    __tablename__ = 'device'

    id = Column(Integer, primary_key=True)
    created_at = Column(DateTime, index=True, nullable=False)
    device_type = Column(Integer, index=True)
    device_sub_type = Column(Integer, index=True)
    device_id = Column(String(20), nullable=False, unique=True)
    area_id = Column(Integer, ForeignKey('area.id'))

    area = relationship(
        "Area", backref=backref('devices'), lazy='joined', uselist=False)

    series = relationship(
        "Series",
        secondary=("join(DeviceSeries, Series, DeviceSeries.series_id == "
                   "Series.id)"),
        primaryjoin=("and_(Device.id == DeviceSeries.device_id, Series.id == "
                     "DeviceSeries.series_id)"),
        secondaryjoin="Series.id == DeviceSeries.series_id"
    )

    def __init__(self, device_type, device_sub_type, device_id, area=None):
        super().__init__()
        self.created_at = datetime.utcnow()
        self.device_type = device_type
        self.device_sub_type = device_sub_type
        self.device_id = device_id
        self.area = area

    @classmethod
    def get_or_create(cls, device_type, device_sub_type, device_id):
        return get_or_create(cls, device_type=device_type,
                             device_sub_type=device_sub_type,
                             device_id=device_id)

    def __repr__(self):
        return "Device(name=%s, ID=%r)" % (
            self.area, self.device_id)



from collections import OrderedDict

from sqlalchemy.sql import text
from flask import render_template, Blueprint

from home.ts.models import db, Area, Device, Series


web = Blueprint('Dashboard Web', __name__)


@web.context_processor
def inject_devices():
    return dict(devices=Device.query.join(Area).order_by(Area.name).all())


@web.route('/')
@web.route('/areas/<device_name>/')
def dashboard(device_name=None):

    series_map = {s.id: s for s in Series.query.all()}

    query = """ SELECT
    device.id as device_id, area.name as device_name,
    series.id as series_id, series.name as series_name,
    ARRAY (
        SELECT hstore(data_point)
        FROM data_point
        WHERE data_point.device_series_id = device_series.id
        ORDER BY data_point.created_at DESC
        LIMIT 30
    ) as latest
    FROM device_series
    JOIN device ON device.id = device_series.device_id
    JOIN series ON series.id = device_series.series_id
    JOIN area ON area.id = device.area_id"""

    if device_name:
        kwargs = {'device_name': device_name}
        query += "\n    WHERE area.name = :device_name"
    else:
        kwargs = {}

    query += "\n    ORDER BY series_name, device_name"

    device_series = db.engine.execute(text(query), **kwargs)

    structured = OrderedDict()

    for row in device_series:

        structured[row.device_name] = structured.get(row.device_name, [])

        structured[row.device_name].append({
            'series_id': row.series_id,
            'model': series_map[row.series_id],
            'device_id': row.device_id,
            'name': row.series_name,
            'value': row.latest[0]['value'],
            'created_at': row.latest[0]['created_at'],
            'data_points': row.latest,
            'sparkline': ','.join(reversed([d['value'] for d in row.latest]))
        })

    if device_name is not None:
        template = 'device.html'
    else:
        template = 'dashboard.html'

    return render_template(
        template,
        device_readings=structured,
        device_name=device_name,
    )



"""Allow devices with no area_id.

Revision ID: 1e665b27760
Revises: 54db49335a2
Create Date: 2014-04-20 16:32:55.982893

"""

# revision identifiers, used by Alembic.
revision = '1e665b27760'
down_revision = '54db49335a2'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.alter_column('device', 'area_id', existing_type=sa.INTEGER(),
                    nullable=True)


def downgrade():
    op.alter_column('device', 'area_id', existing_type=sa.INTEGER(),
                    nullable=False)
